import sys
import json
import logging

import pika

if sys.version_info > (3, 0):
    from PIKACHU.utils.py3 import Single
else:
    from PIKACHU.utils.py2 import Single
from PIKACHU import utils
logger = logging.getLogger(__name__)


"""
Plan: 
1. SimpleProducor/SimpleConsumer use direct exchange_type, message durable and acknowledge is enabled.
   SimpleConsumer is synchronous consumer, SimpleAsyncConsumer is asynchronous consumer
2. Publisher/Subscriber use fanout exchange_type, for a publish/subscribe pattern. 
   Message won't resend if an receiver miss or not ack an message.
3. StreamProducor/StreamConsumer use topic exchange_type, suitable for a topic based multi stream
   subscript pattern (one producor, many consumers with filter).
4. Only SimpleConsumer has a synchronous version, other consumer/receiver are all asynchronous, 
   cause only SimpleProducor keeps a durable message queue, so it's consumer can be "get" off and on,
   while asynchronous consumers/receivers can only "listen" in an ioloop. That's why synchronous consumer
   only has a get method while asynchronous ones only have start_listen method.
"""


class SimpleProducor(Single):
    """
    SimpleProducor use singleton pattern. Connection to rabbitmq is BlockingConnection. All messages are durable and ack needed.
    """
    _connection = None
    _channel = None
    EXCHANGE_TYPE = "direct"

    def __init__(self, url, namespace=None):
        self._url = url
        self._namespace = namespace or "pikachu"
        self.__prepare_channel()

    def __prepare_channel(self):
        self._connection = pika.BlockingConnection(pika.URLParameters(self._url))
        self._channel = self._connection.channel()
        self._channel.confirm_delivery()
        return self._channel

    def __ensure_channel(self):
        if self._connection.is_closed or self._channel.is_closed:
            self.__prepare_channel()
        else:
            # detect server disconnect
            try:
                self._connection.process_data_events()
            except pika.exceptions.ConnectionClosed:
                logger.info("Connection closed by server, try reconnect.")
                self.__prepare_channel()
        return self._channel

    def put(self, message):
        """
        The put/get or put/listen pattern. Put a message into the queue.
        :message: the message to put in queue, dict.
        """
        assert isinstance(message, dict), "Message only support dict"
        body = json.dumps(message)
        exchange = utils.make_exchange_name(self._namespace, self.EXCHANGE_TYPE)
        routing_key = utils.make_direct_key(self._namespace)
        queue_name = utils.make_queue_name(self._namespace, self.EXCHANGE_TYPE)
        channel = self.__ensure_channel()
        channel.exchange_declare(exchange=exchange, exchange_type=self.EXCHANGE_TYPE, durable=True)
        channel.queue_declare(queue=queue_name, durable=True)  # make sure the queue is created.
        binding_key = routing_key  # direct mode only match routing_key with binding_key, they should be the same
        channel.queue_bind(queue_name, exchange, binding_key)
        return channel.basic_publish(
            exchange=exchange,
            routing_key=routing_key,
            body=body,
            properties=pika.BasicProperties(delivery_mode=2,))


# TODO: Publisher, StreamProducor


import sys
sys.path.append("../")
import random

import PIKACHU
import settings

producor = PIKACHU.SimpleProducor(settings.amqp)
message = dict(data=random.randint(0, 666))
print("put message {} result: {}".format(message, producor.put(message)))



import sys
import time
sys.path.append("../")
import PIKACHU
import settings

consumer = PIKACHU.SimpleConsumer(settings.amqp)
envelopes = consumer.get()
print("get {} message".format(len(envelopes)))

for envelope in envelopes:
    print("get message:", envelope.message)
    envelope.message_read()




import sys
import time
import tornado
import tornado.web
sys.path.append("../")
import PIKACHU
import settings


def cb(envelope):
    print("get message:", envelope.message)
    envelope.message_read()


class IndexHandler(tornado.web.RequestHandler):
    def get(self):
        self.write("Test!")

application = tornado.web.Application([
    (r"/", IndexHandler),
])
application.listen(8080)
consumer = PIKACHU.SimpleAsyncConsumer(settings.amqp, tornado_mode=True)
consumer.start_listen(cb)
ioloop = tornado.ioloop.IOLoop.current()
ioloop.start()



import json
import time
import logging

import pika

from PIKACHU import utils

logger = logging.getLogger(__name__)

class Envelope(object):
    def __init__(self, channel, basic_deliver, properties, body):
        self.channel = channel
        self.basic_deliver = basic_deliver
        self.properties = properties
        self.message = json.loads(body)

    def message_read(self):
        """
        acknowledge the queue that the message has been proccessed.
        """
        self.channel.basic_ack(self.basic_deliver.delivery_tag)


class SimpleConsumer(object):
    EXCHANGE_TYPE = "direct"
    def __init__(self, url, namespace=None):
        self._url = url
        self._connection = pika.BlockingConnection(pika.URLParameters(url))
        self._channel = self._connection.channel()
        self._queue_name = utils.make_queue_name(namespace or "pikachu", self.EXCHANGE_TYPE)
        self._channel.queue_declare(self._queue_name, durable=True)
        

    def get(self, max_len=100):
        """
        Get message from queue, 100 messages max by default.
        :max_len: the max message count to get
        :return: list of Envelope
        """
        
        envelopes = []
        for i in range(max_len):
            basic_deliver, properties, body = self._channel.basic_get(self._queue_name)
            if body:
                envelopes.append(Envelope(self._channel, basic_deliver, properties, body))
            else:
                break
        return envelopes
        

class SimpleAsyncConsumer(object):
    _connection = None
    _channel = None
    EXCHANGE_TYPE = "direct"
    def __init__(self, url, namespace=None, tornado_mode=False):
        self._url = url
        self._namespace = namespace or "pikachu"
        self._tornado_mode = tornado_mode
        self.Connection = pika.TornadoConnection if tornado_mode else pika.SelectConnection

    def _connect(self):
        # TODO: handle connect fail exception, try reconnect.
        return self.Connection(
            pika.URLParameters(self._url),
            on_open_callback=self.__on_connection_open,
            on_close_callback=self.__on_connection_close)

    def _reconnect(self, tried_times=0):
        logger.info("Reconnect, times: {}".format(tried_times))
        if self._connection and self._connection.is_open:
            logger.info("Already connected, abort.")
            return
        if tried_times >= 5:
            logger.info("Reconnect fail! Retry too many times.")
            return
        self._connection = self._connect()
        try:
            if not self._tornado_mode:
                self._connection.ioloop.start()
            else:
                self._connection.add_timeout(tried_times*5, lambda : self._reconnect(tried_times=tried_times+1))
        except pika.exceptions.AMQPConnectionError:
            time.sleep(tried_times*5)
            self._reconnect(tried_times=tried_times+1)


    def __on_connection_close(self, connection, reply_code, reply_text):
        self._channel = None
        logger.info('connection {} closed for reason [{}: {}], reconnect in 5s...'.format(connection, reply_code, reply_text))
        if self._tornado_mode:
            self._connection.add_timeout(5, self._reconnect)
        else:
            time.sleep(5)
            self._reconnect()

    def __on_message(self, channel, basic_deliver, properties, body):
        message = Envelope(channel, basic_deliver, properties, body)
        self.callback(message)

    def start_listen(self, callback_on_message):
        """
        :callback: on message callback, callback(Envelope), will pass an Envelope object to the callback.
        :return: the ioloop
        """
        self.callback = callback_on_message
        connection = self._connect()
        return connection.ioloop

    def __on_connection_open(self, connection):
        logger.info("Connected!")
        self._connection = connection
        self._channel = connection.channel(on_open_callback=self.__on_channel_open)

    def __on_channel_closed(self, channel, reply_code, reply_text):
        logger.info("Channel {} is closed for readon [{}:{}]".format(channel, reply_code, reply_text))
        # try reopen channel if connection is still open
        if self._connection and self._connection.is_open:
            self._channel = self._connection.channel(on_open_callback=self.__on_channel_open)

    def __on_channel_open(self, channel):
        self._channel.add_on_close_callback(self.__on_channel_closed)
        # consumer don't need to declare the exchage or bind queue to exchange
        queue_name = utils.make_queue_name(self._namespace, self.EXCHANGE_TYPE)
        self._channel.queue_declare(self.__on_queue_declareok, queue=queue_name, durable=True)

    def __on_queue_declareok(self, method_frame):
        queue_name = utils.make_queue_name(self._namespace, self.EXCHANGE_TYPE)
        self._consume_tag = self._channel.basic_consume(self.__on_message, queue=queue_name, no_ack=False)


def make_exchange_name(namespace, exchange_type):
    return "{}.{}".format(namespace, exchange_type)

def make_channel_name(namespace, exchange_type):
    return "channel_on_{}.{}".format(namespace, exchange_type)

def make_queue_name(namespace, exchange_type):
    return "queue_for_{}.{}".format(namespace, exchange_type)

def make_direct_key(namespace):
    return "key_for_{}.direct".format(namespace)


import sys
import time
sys.path.append("../")
import PIKACHU
import settings

consumer = PIKACHU.SimpleAsyncConsumer(settings.amqp)
def cb(envelope):
    print("get message:", envelope.message)
    envelope.message_read()
ioloop = consumer.start_listen(cb)
ioloop.start()



# -*- coding: utf-8 -*-

import sys
import setuptools
from distutils.core import setup


with open("README.md", "r") as fh:
    long_description = fh.read()

def get_info():
    init_file = 'PIKACHU/__init__.py'
    with open(init_file, 'r') as f:
        for line in f.readlines():
            if "=" in line:
                exec(compile(line, "", 'exec'))
    return locals()['name'], locals()['author'], locals()['version']

NAME, AUTHOR, VERSION = get_info()

sys.dont_write_bytecode = True
setuptools.setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email="fufu.bluesand@gmail.com",
    description="a PIKA based, Cuter and more Human rabbitmq queue Utility (´_ゝ`)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/smilefufu/PIKACHU",
    data_files = [("", ["LICENSE"])],
    packages=setuptools.find_packages(),
    install_requires=[
        "pika",
    ],
    classifiers=(
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: OS Independent'
    ),
)
